@@ -112,7 +112,10 @@ def process_dataset(data, dataset_id=None, metadata=None,
     data_path = None
     if isinstance(data, pandas.DataFrame):
         metadata['nb_rows'] = len(data)
-        data = data.astype(str)
+        # Change to object dtype first and do fillna() to work around bug
+        # https://github.com/pandas-dev/pandas/issues/25353 (nan as str 'nan')
+        data = data.astype(object).fillna('').astype(str)
+
         # FIXME: no sampling here!
     else:
         with contextlib.ExitStack() as stack:
@@ -181,8 +184,9 @@ def process_dataset(data, dataset_id=None, metadata=None,
 
     # Set column names
     for column_meta, name in zip(columns, data.columns):
-        if 'name' not in column_meta:
-            column_meta['name'] = name
+        if 'name' in column_meta and column_meta['name'] != name:
+            raise ValueError("Column names don't match")
+        column_meta['name'] = name
 
     if data.shape[0] == 0:
         logger.info("0 rows, returning early")
@@ -193,7 +197,7 @@ def process_dataset(data, dataset_id=None, metadata=None,
     columns_long = []
 
     # Textual columns
-    column_textual = []
+    columns_textual = []
 
     # Addresses
     resolved_addresses = {}
@@ -201,15 +205,35 @@ def process_dataset(data, dataset_id=None, metadata=None,
     # Administrative areas
     resolved_admin_areas = {}
 
+    # Get manual updates from the user
+    manual_columns = {}
+    manual_latlong_pairs = {}
+    if 'manual_annotations' in metadata:
+        if 'columns' in metadata['manual_annotations']:
+            manual_columns = {
+                col['name']: col
+                for col in metadata['manual_annotations']['columns']
+            }
+            manual_latlong_pairs = {
+                col['name']: col['latlong_pair']
+                for col in metadata['manual_annotations']['columns']
+                if 'latlong_pair' in col
+            }
+
     # Identify types
     logger.info("Identifying types, %d columns...", len(columns))
     with PROM_TYPES.time():
-        for i, column_meta in enumerate(columns):
-            logger.info("Processing column %d...", i)
-            array = data.iloc[:, i]
+        for column_idx, column_meta in enumerate(columns):
+            logger.info("Processing column %d...", column_idx)
+            array = data.iloc[:, column_idx]
             # Identify types
+            if column_meta['name'] in manual_columns:
+                manual = manual_columns[column_meta['name']]
+            else:
+                manual = None
+
             structural_type, semantic_types_dict, additional_meta = \
-                identify_types(array, column_meta['name'], geo_data)
+                identify_types(array, column_meta['name'], geo_data, manual)
             # Set structural type
             column_meta['structural_type'] = structural_type
             # Add semantic types to the ones already present
@@ -259,13 +283,17 @@ def process_dataset(data, dataset_id=None, metadata=None,
 
                 # Get lat/long columns
                 if types.LATITUDE in semantic_types_dict:
-                    columns_lat.append(
-                        (column_meta['name'], numerical_values)
-                    )
+                    columns_lat.append((
+                        column_meta['name'],
+                        numerical_values,
+                        manual_latlong_pairs.get(column_meta['name']),
+                    ))
                 elif types.LONGITUDE in semantic_types_dict:
-                    columns_long.append(
-                        (column_meta['name'], numerical_values)
-                    )
+                    columns_long.append((
+                        column_meta['name'],
+                        numerical_values,
+                        manual_latlong_pairs.get(column_meta['name']),
+                    ))
                 elif coverage:
                     ranges = get_numerical_ranges(
                         [x for x in numerical_values if x is not None]
@@ -364,7 +392,7 @@ def process_dataset(data, dataset_id=None, metadata=None,
                 structural_type == types.TEXT and
                 types.DATE_TIME not in semantic_types_dict
             ):
-                column_textual.append(column_meta['name'])
+                columns_textual.append(column_idx)
 
             # Resolve addresses into coordinates
             if (
@@ -399,10 +427,9 @@ def process_dataset(data, dataset_id=None, metadata=None,
                 resolved_admin_areas[column_meta['name']] = areas
 
     # Textual columns
-    if lazo_client and column_textual:
+    if lazo_client and columns_textual:
         # Indexing with lazo
         if not search:
-            # TODO: Remove previous data from lazo
             logger.info("Indexing textual data with Lazo...")
             try:
                 if data_path:
@@ -410,15 +437,15 @@ def process_dataset(data, dataset_id=None, metadata=None,
                     lazo_client.index_data_path(
                         data_path,
                         dataset_id,
-                        column_textual
+                        [columns[idx]['name'] for idx in columns_textual],
                     )
                 else:
                     # if path is not available, send the data instead
-                    for column_name in column_textual:
+                    for idx in columns_textual:
                         lazo_client.index_data(
-                            data[column_name].values.tolist(),
+                            data.iloc[:, idx].values.tolist(),
                             dataset_id,
-                            column_name
+                            columns[idx]['name'],
                         )
             except Exception:
                 logger.error("Error indexing textual attributes from %s", dataset_id)
@@ -432,24 +459,23 @@ def process_dataset(data, dataset_id=None, metadata=None,
                     lazo_sketches = lazo_client.get_lazo_sketch_from_data_path(
                         data_path,
                         "",
-                        column_textual,
+                        [columns[idx]['name'] for idx in columns_textual],
                     )
                 else:
                     # if path is not available, send the data instead
                     lazo_sketches = []
-                    for column_name in column_textual:
+                    for idx in columns_textual:
                         lazo_sketches.append(
                             lazo_client.get_lazo_sketch_from_data(
-                                data[column_name].values.tolist(),
+                                data.iloc[:, idx].values.tolist(),
                                 "",
-                                column_name,
+                                columns[idx]['name'],
                             )
                         )
                 # saving sketches into metadata
-                for i in range(len(column_textual)):
-                    n_permutations, hash_values, cardinality = \
-                        lazo_sketches[i]
-                    columns[i]['lazo'] = dict(
+                for sketch, idx in zip(lazo_sketches, columns_textual):
+                    n_permutations, hash_values, cardinality = sketch
+                    columns[idx]['lazo'] = dict(
                         n_permutations=n_permutations,
                         hash_values=list(hash_values),
                         cardinality=cardinality,
@@ -481,7 +507,7 @@ def process_dataset(data, dataset_id=None, metadata=None,
                     col['semantic_types'].remove(types.LONGITUDE)
 
             # Compute ranges from lat/long pairs
-            for (name_lat, values_lat), (name_long, values_long) in pairs:
+            for (name_lat, values_lat, annot_pair), (name_long, values_long, annot_pair) in pairs:
                 values = []
                 for lat, long in zip(values_lat, values_long):
                     if (lat and long and  # Ignore None and 0
@@ -576,7 +602,7 @@ def process_dataset(data, dataset_id=None, metadata=None,
         choose_rows.sort()  # Keep it in order
         sample = data.iloc[choose_rows]
         sample = sample.applymap(truncate_string)  # Truncate long values
-        metadata['sample'] = sample.to_csv(index=False)
+        metadata['sample'] = sample.to_csv(index=False, line_terminator='\r\n')
 
     # Return it -- it will be inserted into Elasticsearch, and published to the
     # feed and the waiting on-demand searches
